MODULE icethd_zdf_BL99
   !!======================================================================
   !!                       ***  MODULE icethd_zdf_BL99 ***
   !!   sea-ice: vertical heat diffusion in sea ice (computation of temperatures) 
   !!======================================================================
   !! History :       !  2003-02  (M. Vancoppenolle) original 1D code
   !!                 !  2005-06  (M. Vancoppenolle) 3d version
   !!            4.0  !  2018     (many people)      SI3 [aka Sea Ice cube]
   !!----------------------------------------------------------------------
#if defined key_si3
   !!----------------------------------------------------------------------
   !!   'key_si3'                                       SI3 sea-ice model
   !!----------------------------------------------------------------------
   !!  ice_thd_zdf_BL99 : vertical diffusion computation
   !!----------------------------------------------------------------------
   USE dom_oce        ! ocean space and time domain
   USE phycst         ! physical constants (ocean directory) 
   USE ice            ! sea-ice: variables
   USE ice1D          ! sea-ice: thermodynamics variables
   USE icevar         ! sea-ice: operations
   !
   USE in_out_manager ! I/O manager
   USE lib_mpp        ! MPP library
   USE lib_fortran    ! fortran utilities (glob_sum + no signed zero)

   IMPLICIT NONE
   PRIVATE

   PUBLIC   ice_thd_zdf_BL99   ! called by icethd_zdf

   !!----------------------------------------------------------------------
   !! NEMO/ICE 4.0 , NEMO Consortium (2018)
   !! $Id$
   !! Software governed by the CeCILL license (see ./LICENSE)
   !!----------------------------------------------------------------------
CONTAINS

   SUBROUTINE ice_thd_zdf_BL99( k_jules )
      !!-------------------------------------------------------------------
      !!                ***  ROUTINE ice_thd_zdf_BL99  ***
      !!
      !! ** Purpose : computes the time evolution of snow and sea-ice temperature
      !!              profiles, using the original Bitz and Lipscomb (1999) algorithm
      !!
      !! ** Method  : solves the heat equation diffusion with a Neumann boundary
      !!              condition at the surface and a Dirichlet one at the bottom. 
      !!              Solar radiation is partially absorbed into the ice.
      !!              The specific heat and thermal conductivities depend on ice 
      !!              salinity and temperature to take into account brine pocket   
      !!              melting. The numerical scheme is an iterative Crank-Nicolson
      !!              on a non-uniform multilayer grid in the ice and snow system.
      !!
      !!           The successive steps of this routine are
      !!           1.  initialization of ice-snow layers thicknesses
      !!           2.  Internal absorbed and transmitted radiation
      !!           Then iterative procedure begins
      !!           3.  Thermal conductivity
      !!           4.  Kappa factors
      !!           5.  specific heat in the ice
      !!           6.  eta factors
      !!           7.  surface flux computation
      !!           8.  tridiagonal system terms
      !!           9.  solving the tridiagonal system with Gauss elimination
      !!           Iterative procedure ends according to a criterion on evolution
      !!           of temperature
      !!           10. Fluxes at the interfaces
      !!
      !! ** Inputs / Ouputs : (global commons)
      !!           surface temperature              : t_su_1d
      !!           ice/snow temperatures            : t_i_1d, t_s_1d
      !!           ice salinities                   : sz_i_1d
      !!           number of layers in the ice/snow : nlay_i, nlay_s
      !!           total ice/snow thickness         : h_i_1d, h_s_1d
      !!-------------------------------------------------------------------
      INTEGER, INTENT(in) ::   k_jules     ! Jules coupling (0=OFF, 1=EMULATED, 2=ACTIVE)
      !
      INTEGER ::   ji, jk         ! spatial loop index
      INTEGER ::   jm             ! current reference number of equation
      INTEGER ::   jm_mint, jm_maxt
      INTEGER ::   iconv          ! number of iterations in iterative procedure
      INTEGER ::   iconv_max = 50 ! max number of iterations in iterative procedure
      !
      INTEGER, DIMENSION(jpij) ::   jm_min    ! reference number of top equation
      INTEGER, DIMENSION(jpij) ::   jm_max    ! reference number of bottom equation
      !
      REAL(wp) ::   zg1s      =  2._wp        ! for the tridiagonal system
      REAL(wp) ::   zg1       =  2._wp        !
      REAL(wp) ::   zgamma    =  18009._wp    ! for specific heat
      REAL(wp) ::   zbeta     =  0.117_wp     ! for thermal conductivity (could be 0.13)
      REAL(wp) ::   zraext_s  =  10._wp       ! extinction coefficient of radiation in the snow
      REAL(wp) ::   zkimin    =  0.10_wp      ! minimum ice thermal conductivity
      REAL(wp) ::   ztsu_err  =  1.e-5_wp     ! range around which t_su is considered at 0C 
      REAL(wp) ::   zdti_bnd  =  1.e-4_wp     ! maximal authorized error on temperature 
      REAL(wp) ::   zhs_min   =  0.01_wp      ! minimum snow thickness for conductivity calculation 
      REAL(wp) ::   ztmelts                   ! ice melting temperature
      REAL(wp) ::   zdti_max                  ! current maximal error on temperature 
      REAL(wp) ::   zcpi                      ! Ice specific heat
      REAL(wp) ::   zhfx_err, zdq             ! diag errors on heat
      REAL(wp) ::   zfac                      ! dummy factor
      !
      REAL(wp), DIMENSION(jpij) ::   isnow        ! switch for presence (1) or absence (0) of snow
      REAL(wp), DIMENSION(jpij) ::   ztsub        ! surface temperature at previous iteration
      REAL(wp), DIMENSION(jpij) ::   zh_i, z1_h_i ! ice layer thickness
      REAL(wp), DIMENSION(jpij) ::   zh_s, z1_h_s ! snow layer thickness
      REAL(wp), DIMENSION(jpij) ::   zqns_ice_b   ! solar radiation absorbed at the surface
      REAL(wp), DIMENSION(jpij) ::   zfnet        ! surface flux function
      REAL(wp), DIMENSION(jpij) ::   zdqns_ice_b  ! derivative of the surface flux function
      !
      REAL(wp), DIMENSION(jpij       )     ::   ztsuold     ! Old surface temperature in the ice
      REAL(wp), DIMENSION(nlay_i, jpij)     ::   ztiold      ! Old temperature in the ice
      REAL(wp), DIMENSION(nlay_s, jpij)     ::   ztsold      ! Old temperature in the snow
      REAL(wp), DIMENSION(nlay_i, jpij)     ::   ztib        ! Temporary temperature in the ice to check the convergence
      REAL(wp), DIMENSION(nlay_s, jpij)     ::   ztsb        ! Temporary temperature in the snow to check the convergence
      REAL(wp), DIMENSION(0:nlay_i, jpij)   ::   ztcond_i    ! Ice thermal conductivity
      REAL(wp), DIMENSION(jpij, 0:nlay_i)   ::   zradtr_i    ! Radiation transmitted through the ice
      REAL(wp), DIMENSION(0:nlay_i, jpij)   ::   zradab_i    ! Radiation absorbed in the ice
      REAL(wp), DIMENSION(0:nlay_i, jpij)   ::   zkappa_i    ! Kappa factor in the ice
      REAL(wp), DIMENSION(0:nlay_i, jpij)   ::   zeta_i      ! Eta factor in the ice
      REAL(wp), DIMENSION(jpij,0:nlay_s)   ::   zradtr_s    ! Radiation transmited through the snow
      REAL(wp), DIMENSION(0:nlay_s, jpij)   ::   zradab_s    ! Radiation absorbed in the snow
      REAL(wp), DIMENSION(0:nlay_s, jpij)   ::   zkappa_s    ! Kappa factor in the snow
      REAL(wp), DIMENSION(0:nlay_s, jpij)   ::   zeta_s      ! Eta factor in the snow
      REAL(wp), DIMENSION(nlay_i+3, jpij)   ::   zindterm    ! 'Ind'ependent term
      REAL(wp), DIMENSION(nlay_i+3, jpij)   ::   zindtbis    ! Temporary 'ind'ependent term
      REAL(wp), DIMENSION(nlay_i+3, jpij)   ::   zdiagbis    ! Temporary 'dia'gonal term
      REAL(wp), DIMENSION(3, nlay_i+3, jpij)::   ztrid       ! Tridiagonal system terms
      REAL(wp), DIMENSION(jpij)            ::   zq_ini      ! diag errors on heat
      REAL(wp), DIMENSION(jpij)            ::   zghe        ! G(he), th. conduct enhancement factor, mono-cat
      REAL(wp), DIMENSION(nlay_i, jpij)     ::   tt_i_1d, tsz_i_1d
      REAL(wp), DIMENSION(nlay_s, jpij)     ::   tt_s_1d
      !
      ! Mono-category
      REAL(wp) ::   zepsilon   ! determines thres. above which computation of G(h) is done
      REAL(wp) ::   zhe        ! dummy factor
      REAL(wp) ::   zcnd_i     ! mean sea ice thermal conductivity
      INTEGER  ::   itot
      LOGICAL, DIMENSION(jpij)  ::   liter
      !!------------------------------------------------------------------     
      itot = npti
      ! --- diag error on heat diffusion - PART 1 --- !
      DO ji = 1, npti
         zq_ini(ji) = ( SUM( e_i_1d(ji,1:nlay_i) ) * h_i_1d(ji) * r1_nlay_i +  &
            &           SUM( e_s_1d(ji,1:nlay_s) ) * h_s_1d(ji) * r1_nlay_s ) 
      END DO

      !------------------
      ! 1) Initialization
      !------------------
      DO ji = 1, npti
         isnow(ji) = 1._wp - MAX( 0._wp , SIGN(1._wp, - h_s_1d(ji) ) )  ! is there snow or not
         ! layer thickness
         zh_i(ji) = h_i_1d(ji) * r1_nlay_i
         zh_s(ji) = h_s_1d(ji) * r1_nlay_s
      END DO
      !
      WHERE( zh_i(1:npti) >= epsi10 )   ;   z1_h_i(1:npti) = 1._wp / zh_i(1:npti)
      ELSEWHERE                         ;   z1_h_i(1:npti) = 0._wp
      END WHERE
      !
      WHERE( zh_s(1:npti) > 0._wp   )       zh_s(1:npti) = MAX( zhs_min * r1_nlay_s, zh_s(1:npti) )
      !
      WHERE( zh_s(1:npti) > 0._wp   )   ;   z1_h_s(1:npti) = 1._wp / zh_s(1:npti)
      ELSEWHERE                         ;   z1_h_s(1:npti) = 0._wp
      END WHERE
      !
      ! Store initial temperatures and non solar heat fluxes
      IF( k_jules == np_jules_OFF .OR. k_jules == np_jules_EMULE ) THEN
         !
         ztsub      (1:npti) = t_su_1d(1:npti)                          ! surface temperature at iteration n-1
         ztsuold    (1:npti) = t_su_1d(1:npti)                          ! surface temperature initial value
         t_su_1d    (1:npti) = MIN( t_su_1d(1:npti), rt0 - ztsu_err )   ! required to leave the choice between melting or not
         zdqns_ice_b(1:npti) = dqns_ice_1d(1:npti)                      ! derivative of incoming nonsolar flux 
         zqns_ice_b (1:npti) = qns_ice_1d(1:npti)                       ! store previous qns_ice_1d value
         !
      ENDIF
      !
      DO jk = 1, nlay_i
         DO ji = 1, npti
            tt_i_1d(jk, ji) = t_i_1d(ji, jk)
            tsz_i_1d(jk, ji) = sz_i_1d(ji, jk)
         END DO
      END DO

      DO jk = 1, nlay_s
         DO ji = 1, npti
            tt_s_1d(jk, ji) = t_s_1d(ji, jk)
         END DO
      END DO


      ztsold (:, :) = tt_s_1d(:,:)   ! Old snow temperature
      ztiold (:, :) = tt_i_1d(:,:)   ! Old ice temperature

      !-------------
      ! 2) Radiation
      !-------------
      ! --- Transmission/absorption of solar radiation in the ice --- !
      zradtr_s(1:npti,0) = qtr_ice_top_1d(1:npti)
      DO jk = 1, nlay_s
         DO ji = 1, npti
            !                             ! radiation transmitted below the layer-th snow layer
            zradtr_s(ji,jk) = zradtr_s(ji,0) * EXP( - zraext_s * h_s_1d(ji) * r1_nlay_s * REAL(jk) )
            !                             ! radiation absorbed by the layer-th snow layer
            zradab_s(jk,ji) = zradtr_s(ji,jk-1) - zradtr_s(ji,jk)
         END DO
      END DO
      !
      zradtr_i(1:npti, 0) = zradtr_s(1:npti, nlay_s) * isnow(1:npti) + qtr_ice_top_1d(1:npti) * ( 1._wp - isnow(1:npti) )
      DO jk = 1, nlay_i 
         DO ji = 1, npti
            !                             ! radiation transmitted below the layer-th ice layer
            zradtr_i(ji, jk) = zradtr_i(ji, 0) * EXP( - rn_kappa_i * zh_i(ji) * REAL(jk) )
            !                             ! radiation absorbed by the layer-th ice layer
            zradab_i(jk,ji) = zradtr_i(ji,jk-1) - zradtr_i(ji,jk)
         END DO
      END DO
      !
      qtr_ice_bot_1d(1:npti) = zradtr_i(1:npti, nlay_i)   ! record radiation transmitted below the ice
      !
      iconv    = 0          ! number of iterations
      zdti_max = 1000._wp   ! maximal value of error on all points
      !
      liter(:) = .TRUE.                                          !============================!
      DO iconv = 1, iconv_max                                    ! Iterative procedure begins !
         IF(itot<1) EXIT
         DO ji = 1, npti
          IF(liter(ji)) THEN
            zdti_max = 0._wp
            !
            ztib(:, ji) = tt_i_1d(:, ji)
            ztsb(:, ji) = tt_s_1d(:, ji)
            !
            !--------------------------------
            ! 3) Sea ice thermal conductivity
            !--------------------------------
            IF( ln_cndi_U64 ) THEN         !-- Untersteiner (1964) formula: k = k0 + beta.S/T
            !
               ztcond_i(0, ji)      = rcnd_i + zbeta * tsz_i_1d(1, ji)      / MIN( -epsi10, tt_i_1d(1, ji) - rt0 )
               ztcond_i(nlay_i, ji) = rcnd_i + zbeta * tsz_i_1d(nlay_i, ji) / MIN( -epsi10, t_bo_1d(ji)  - rt0 )
               DO jk = 1, nlay_i-1
                    ztcond_i(jk, ji) = rcnd_i + zbeta * 0.5_wp * ( tsz_i_1d(jk, ji) + tsz_i_1d(jk+1, ji) ) /  &
                        &                       MIN( -epsi10, 0.5_wp * (tt_i_1d(jk, ji) + tt_i_1d(jk+1, ji)) - rt0 )
                  END DO
               !
            ELSEIF( ln_cndi_P07 ) THEN     !-- Pringle et al formula: k = k0 + beta1.S/T - beta2.T
               !
                  ztcond_i(0, ji)      = rcnd_i + 0.09_wp  *  tsz_i_1d(1, ji)      / MIN( -epsi10, tt_i_1d(1, ji) - rt0 )  &
                     &                         - 0.011_wp * ( tt_i_1d(1, ji) - rt0 )
                  ztcond_i(nlay_i, ji) = rcnd_i + 0.09_wp  *  tsz_i_1d(nlay_i, ji) / MIN( -epsi10, t_bo_1d(ji)  - rt0 )  &
                     &                         - 0.011_wp * ( t_bo_1d(ji) - rt0 )
               DO jk = 1, nlay_i-1
                  ztcond_i(jk,ji) = rcnd_i + 0.09_wp  *   0.5_wp * ( tsz_i_1d(jk,ji) + tsz_i_1d(jk+1,ji) ) /        &
                     &                      MIN( -epsi10, 0.5_wp * ( tt_i_1d (jk,ji) + tt_i_1d (jk+1,ji) ) - rt0 )  &
                     &                     - 0.011_wp * ( 0.5_wp * ( tt_i_1d (jk,ji) + tt_i_1d (jk+1,ji) ) - rt0 )
               END DO
            !
            ENDIF
            ztcond_i(:, ji) = MAX( zkimin, ztcond_i(:, ji) )        
            !
            !--- G(he) : enhancement of thermal conductivity in mono-category case
            ! Computation of effective thermal conductivity G(h)
            ! Used in mono-category case only to simulate an ITD implicitly
            ! Fichefet and Morales Maqueda, JGR 1997
            zghe(ji) = 1._wp
            !
            SELECT CASE ( nn_virtual_itd )
            !
            CASE ( 1 , 2 )
            !
            zepsilon = 0.1_wp
               zcnd_i = SUM( ztcond_i(:,ji) ) / REAL( nlay_i+1, wp )                                ! Mean sea ice thermal conductivity
               zhe = ( rn_cnd_s * h_i_1d(ji) + zcnd_i * h_s_1d(ji) ) / ( rn_cnd_s + zcnd_i )        ! Effective thickness he (zhe)
               IF( zhe >=  zepsilon * 0.5_wp * EXP(1._wp) )  &
                  &   zghe(ji) = MIN( 2._wp, 0.5_wp * ( 1._wp + LOG( 2._wp * zhe / zepsilon ) ) )   ! G(he)
            !
            END SELECT
            !
            !-----------------
            ! 4) kappa factors
            !-----------------
            !--- Snow
            DO jk = 0, nlay_s-1
               zkappa_s(jk,ji) = zghe(ji) * rn_cnd_s * z1_h_s(ji)
            END DO
            ! Snow-ice interface
            zfac = 0.5_wp * ( ztcond_i(0, ji) * zh_s(ji) + rn_cnd_s * zh_i(ji) )
            IF( zfac > epsi10 ) THEN
               zkappa_s(nlay_s, ji) = zghe(ji) * rn_cnd_s * ztcond_i(0, ji) / zfac
            ELSE
               zkappa_s(nlay_s, ji) = 0._wp
            ENDIF

            !--- Ice
            DO jk = 0, nlay_i
               zkappa_i(jk,ji) = zghe(ji) * ztcond_i(jk,ji) * z1_h_i(ji)
            END DO
            zkappa_i(0, ji) = zkappa_s(nlay_s, ji) * isnow(ji) + zkappa_i(0, ji) * ( 1._wp - isnow(ji) )
            !
            !--------------------------------------
            ! 5) Sea ice specific heat, eta factors
            !--------------------------------------
            DO jk = 1, nlay_i
               zcpi = rcpi + zgamma * tsz_i_1d(jk,ji) / MAX( ( tt_i_1d(jk,ji) - rt0 ) * ( ztiold(jk,ji) - rt0 ), epsi10 )
               zeta_i(jk,ji) = rdt_ice * r1_rhoi * z1_h_i(ji) / MAX( epsi10, zcpi )
            END DO

            DO jk = 1, nlay_s
               zeta_s(jk,ji) = rdt_ice * r1_rhos * r1_rcpi * z1_h_s(ji)
            END DO
         !
         !----------------------------------------!
         !                                        !
         !   JULES COUPLING IS OFF OR EMULATED    !
         !                                        !
         !----------------------------------------!
         !
         IF( k_jules == np_jules_OFF .OR. k_jules == np_jules_EMULE ) THEN
            !
            ! ==> The original BL99 temperature computation is used
            !       (with qsr_ice, qns_ice and dqns_ice as inputs)
            !
            !----------------------------
            ! 6) surface flux computation
            !----------------------------
            ! update of the non solar flux according to the update in T_su
               qns_ice_1d(ji) = qns_ice_1d(ji) + dqns_ice_1d(ji) * ( t_su_1d(ji) - ztsub(ji) )

               zfnet(ji) = qsr_ice_1d(ji) - qtr_ice_top_1d(ji) + qns_ice_1d(ji) ! net heat flux = net - transmitted solar + non solar
            !
            !----------------------------
            ! 7) tridiagonal system terms
            !----------------------------
            ! layer denotes the number of the layer in the snow or in the ice
            ! jm denotes the reference number of the equation in the tridiagonal
            ! system, terms of tridiagonal system are indexed as following :
            ! 1 is subdiagonal term, 2 is diagonal and 3 is superdiagonal one

            ! ice interior terms (top equation has the same form as the others)
            ztrid   (:, :, ji) = 0._wp
            zindterm(:, ji)   = 0._wp
            zindtbis(:, ji)   = 0._wp
            zdiagbis(:, ji)   = 0._wp

            DO jm = nlay_s + 2, nlay_s + nlay_i 
                  jk = jm - nlay_s - 1
                  ztrid   (1, jm, ji) =       - zeta_i(jk,ji) *   zkappa_i(jk-1, ji)
                  ztrid   (2, jm, ji) = 1._wp + zeta_i(jk,ji) * ( zkappa_i(jk-1, ji) + zkappa_i(jk, ji) )
                  ztrid   (3, jm, ji) =       - zeta_i(jk,ji) *                       zkappa_i(jk, ji)
                  zindterm(jm,ji)   = ztiold(jk,ji) + zeta_i(jk,ji) * zradab_i(jk, ji)
            END DO

            jm =  nlay_s + nlay_i + 1
               ! ice bottom term
               ztrid   (1, jm, ji) =       - zeta_i(nlay_i, ji) *   zkappa_i(nlay_i-1, ji)   
               ztrid   (2, jm, ji) = 1._wp + zeta_i(nlay_i, ji) * ( zkappa_i(nlay_i-1, ji) + zkappa_i(nlay_i, ji) * zg1 )
               ztrid   (3, jm, ji) = 0._wp
               zindterm(jm,ji)   = ztiold(nlay_i, ji) + zeta_i(nlay_i, ji) *  &
                  &              ( zradab_i(nlay_i, ji) + zkappa_i(nlay_i, ji) * zg1 * t_bo_1d(ji) ) 

               !                               !---------------------!
               IF( h_s_1d(ji) > 0._wp ) THEN   !  snow-covered cells !
                  !                            !---------------------!
                  ! snow interior terms (bottom equation has the same form as the others)
                  DO jm = 3, nlay_s + 1
                     jk = jm - 1
                     ztrid   (1, jm, ji) =       - zeta_s(jk,ji) *   zkappa_s(jk-1,ji)
                     ztrid   (2, jm, ji) = 1._wp + zeta_s(jk,ji) * ( zkappa_s(jk-1,ji) + zkappa_s(jk,ji) )
                     ztrid   (3, jm, ji) =       - zeta_s(jk,ji) *                       zkappa_s(jk,ji)
                     zindterm(jm,ji)   = ztsold(jk,ji) + zeta_s(jk,ji) * zradab_s(jk,ji)
                  END DO
                  
                  ! case of only one layer in the ice (ice equation is altered)
                  IF( nlay_i == 1 ) THEN
                     ztrid   (3, nlay_s+2, ji) = 0._wp
                     zindterm(nlay_s+2, ji)   = zindterm(nlay_s+2, ji) + zeta_i(1, ji) * zkappa_i(1, ji) * t_bo_1d(ji) 
                  ENDIF
                  
                  IF( t_su_1d(ji) < rt0 ) THEN   !--  case 1 : no surface melting
                     
                     jm_min(ji) = 1
                     jm_max(ji) = nlay_i + nlay_s + 1
                     
                     ! surface equation
                     ztrid   (1, 1, ji) = 0._wp
                     ztrid   (2, 1, ji) = zdqns_ice_b(ji) - zg1s * zkappa_s(0, ji)
                     ztrid   (3, 1, ji) =                   zg1s * zkappa_s(0, ji)
                     zindterm(1, ji)   = zdqns_ice_b(ji) * t_su_1d(ji) - zfnet(ji)
                     
                     ! first layer of snow equation
                     ztrid   (1, 2, ji) =       - zeta_s(1, ji) *                    zkappa_s(0, ji) * zg1s
                     ztrid   (2, 2, ji) = 1._wp + zeta_s(1, ji) * ( zkappa_s(1, ji) + zkappa_s(0, ji) * zg1s )
                     ztrid   (3, 2, ji) =       - zeta_s(1, ji) *   zkappa_s(1, ji)
                     zindterm(2, ji)   = ztsold(1, ji) + zeta_s(1, ji) * zradab_s(1, ji)
                     
                  ELSE                            !--  case 2 : surface is melting
                     !
                     jm_min(ji) = 2
                     jm_max(ji) = nlay_i + nlay_s + 1
                     
                     ! first layer of snow equation
                     ztrid   (1, 2, ji) = 0._wp
                     ztrid   (2, 2, ji) = 1._wp + zeta_s(1, ji) * ( zkappa_s(1, ji) + zkappa_s(0, ji) * zg1s )
                     ztrid   (3, 2, ji) =       - zeta_s(1, ji) *   zkappa_s(1, ji) 
                     zindterm(2, ji)   = ztsold(1, ji) + zeta_s(1, ji) * ( zradab_s(1, ji) + zkappa_s(0, ji) * zg1s * t_su_1d(ji) ) 
                  ENDIF
                  !                            !---------------------!
               ELSE                            ! cells without snow  !
                  !                            !---------------------!
                  !
                  IF( t_su_1d(ji) < rt0 ) THEN   !--  case 1 : no surface melting
                     !
                     jm_min(ji) = nlay_s + 1
                     jm_max(ji) = nlay_i + nlay_s + 1
                     
                     ! surface equation	
                     ztrid   (1, jm_min(ji), ji) = 0._wp
                     ztrid   (2, jm_min(ji), ji) = zdqns_ice_b(ji) - zkappa_i(0, ji) * zg1    
                     ztrid   (3, jm_min(ji), ji) =                   zkappa_i(0, ji) * zg1
                     zindterm(jm_min(ji), ji)   = zdqns_ice_b(ji) * t_su_1d(ji) - zfnet(ji)
                     
                     ! first layer of ice equation
                     ztrid   (1, jm_min(ji)+1, ji) =       - zeta_i(1, ji) *                    zkappa_i(0, ji) * zg1
                     ztrid   (2, jm_min(ji)+1, ji) = 1._wp + zeta_i(1, ji) * ( zkappa_i(1, ji) + zkappa_i(0, ji) * zg1 )
                     ztrid   (3, jm_min(ji)+1, ji) =       - zeta_i(1, ji) *   zkappa_i(1, ji)  
                     zindterm(jm_min(ji)+1, ji)   = ztiold(1, ji) + zeta_i(1, ji) * zradab_i(1, ji)  
                     
                     ! case of only one layer in the ice (surface & ice equations are altered)
                     IF( nlay_i == 1 ) THEN
                        ztrid   (1, jm_min(ji), ji)   = 0._wp
                        ztrid   (2, jm_min(ji), ji)   = zdqns_ice_b(ji)      -   zkappa_i(0, ji) * 2._wp
                        ztrid   (3, jm_min(ji), ji)   =                          zkappa_i(0, ji) * 2._wp
                        ztrid   (1, jm_min(ji)+1, ji) =       - zeta_i(1, ji) *   zkappa_i(0, ji) * 2._wp
                        ztrid   (2, jm_min(ji)+1, ji) = 1._wp + zeta_i(1, ji) * ( zkappa_i(0, ji) * 2._wp + zkappa_i(1, ji) )
                        ztrid   (3, jm_min(ji)+1, ji) = 0._wp
                        zindterm(jm_min(ji)+1, ji)   = ztiold(1, ji) + zeta_i(1, ji) * (zradab_i(1, ji) + zkappa_i(1, ji) * t_bo_1d(ji))
                     ENDIF
                     
                  ELSE                            !--  case 2 : surface is melting
                     
                     jm_min(ji) = nlay_s + 2
                     jm_max(ji) = nlay_i + nlay_s + 1
                     
                     ! first layer of ice equation
                     ztrid   (1, jm_min(ji), ji) = 0._wp
                     ztrid   (2, jm_min(ji), ji) = 1._wp + zeta_i(1, ji) * ( zkappa_i(1, ji) + zkappa_i(0, ji) * zg1 )  
                     ztrid   (3, jm_min(ji), ji) =       - zeta_i(1, ji) *   zkappa_i(1, ji)
                     zindterm(jm_min(ji), ji)   = ztiold(1, ji) + zeta_i(1, ji) * (zradab_i(1, ji) + zkappa_i(0, ji) * zg1 * t_su_1d(ji)) 
                     
                     ! case of only one layer in the ice (surface & ice equations are altered)
                     IF( nlay_i == 1 ) THEN
                        ztrid   (1, jm_min(ji), ji) = 0._wp
                        ztrid   (2, jm_min(ji), ji) = 1._wp + zeta_i(1, ji) * ( zkappa_i(0, ji) * 2._wp + zkappa_i(1, ji) )
                        ztrid   (3, jm_min(ji), ji) = 0._wp
                        zindterm(jm_min(ji), ji)   = ztiold(1, ji) + zeta_i(1, ji) * ( zradab_i(1, ji) + zkappa_i(1, ji) * t_bo_1d(ji) ) &
                           &                      + t_su_1d(ji) * zeta_i(1, ji) * zkappa_i(0, ji) * 2._wp
                     ENDIF
                     
                  ENDIF
               ENDIF
               !
               zindtbis(jm_min(ji), ji) = zindterm(jm_min(ji), ji)
               zdiagbis(jm_min(ji), ji) = ztrid   (2, jm_min(ji), ji)
               !
            !
            !------------------------------
            ! 8) tridiagonal system solving
            !------------------------------
            ! Solve the tridiagonal system with Gauss elimination method.
            ! Thomas algorithm, from Computational fluid Dynamics, J.D. ANDERSON, McGraw-Hill 1984





               DO jm = jm_min(ji)+1, jm_max(ji)
                  zdiagbis(jm,ji) = ztrid   (2, jm, ji) - ztrid(1, jm, ji) * ztrid   (3, jm-1, ji) / zdiagbis(jm-1,ji)
                  zindtbis(jm,ji) = zindterm(jm,ji  ) - ztrid(1, jm, ji) * zindtbis(jm-1,ji  ) / zdiagbis(jm-1,ji)
               END DO

            ! ice temperatures
               tt_i_1d(nlay_i, ji) = zindtbis(jm_max(ji), ji) / zdiagbis(jm_max(ji), ji)

               DO jm = nlay_i + nlay_s, nlay_s + 2, -1
                  jk = jm - nlay_s - 1
                  tt_i_1d(jk, ji) = ( zindtbis(jm,ji) - ztrid(3, jm,ji) * tt_i_1d(jk+1, ji) ) / zdiagbis(jm,ji)
               END DO

               ! snow temperatures      
               IF( h_s_1d(ji) > 0._wp ) THEN
                  tt_s_1d(nlay_s, ji) = ( zindtbis(nlay_s+1, ji) - ztrid(3, nlay_s+1, ji) * tt_i_1d(1, ji) ) / zdiagbis(nlay_s+1, ji)
               ENDIF
               ! surface temperature
               ztsub(ji) = t_su_1d(ji)
               IF( t_su_1d(ji) < rt0 ) THEN
                  t_su_1d(ji) = ( zindtbis(jm_min(ji), ji) - ztrid(3,jm_min(ji),ji) *  &
                     &          ( isnow(ji) * tt_s_1d(1, ji) + ( 1._wp - isnow(ji) ) * tt_i_1d(1, ji) ) ) / zdiagbis(jm_min(ji), ji)
               ENDIF
            !
            !--------------------------------------------------------------
            ! 9) Has the scheme converged?, end of the iterative procedure
            !--------------------------------------------------------------
            ! check that nowhere it has started to melt
            ! zdti_max is a measure of error, it has to be under zdti_bnd

               t_su_1d(ji) = MAX( MIN( t_su_1d(ji) , rt0 ) , rt0 - 100._wp )
               zdti_max    = MAX( zdti_max, ABS( t_su_1d(ji) - ztsub(ji) ) )     

               DO jk = 1, nlay_s
                  tt_s_1d(jk,ji) = MAX( MIN( tt_s_1d(jk,ji), rt0 ), rt0 - 100._wp )
                  zdti_max      = MAX( zdti_max, ABS( tt_s_1d(jk,ji) - ztsb(jk,ji) ) )
               END DO

               DO jk = 1, nlay_i
                  ztmelts       = -rTmlt * tsz_i_1d(jk, ji) + rt0 
                  tt_i_1d(jk, ji) =  MAX( MIN( tt_i_1d(jk, ji), ztmelts ), rt0 - 100._wp )
                  zdti_max      =  MAX( zdti_max, ABS( tt_i_1d(jk, ji) - ztib(jk,ji) ) )
               END DO
         !
         !----------------------------------------!
         !                                        !
         !      JULES COUPLING IS ACTIVE          !
         !                                        !
         !----------------------------------------!
         !
         ELSEIF( k_jules == np_jules_ACTIVE ) THEN
            !
            ! ==> we use a modified BL99 solver with conduction flux (qcn_ice) as forcing term
            !
            !----------------------------
            ! 7) tridiagonal system terms
            !----------------------------
            ! layer denotes the number of the layer in the snow or in the ice
            ! jm denotes the reference number of the equation in the tridiagonal
            ! system, terms of tridiagonal system are indexed as following :
            ! 1 is subdiagonal term, 2 is diagonal and 3 is superdiagonal one

            ! ice interior terms (top equation has the same form as the others)
               ztrid   (:, :, ji) = 0._wp
               zindterm(:, ji)   = 0._wp
               zindtbis(:, ji)   = 0._wp
               zdiagbis(:, ji)   = 0._wp

               DO jm = nlay_s + 2, nlay_s + nlay_i 
                  jk = jm - nlay_s - 1
                  ztrid   (1, jm, ji) =       - zeta_i(jk,ji) *   zkappa_i(jk-1, ji)
                  ztrid   (2, jm, ji) = 1._wp + zeta_i(jk,ji) * ( zkappa_i(jk-1, ji) + zkappa_i(jk,ji) )
                  ztrid   (3, jm, ji) =       - zeta_i(jk,ji) *                       zkappa_i(jk,ji)
                  zindterm(jm,ji)   = ztiold(jk,ji) + zeta_i(jk,ji) * zradab_i(jk,ji)
               ENDDO

               jm =  nlay_s + nlay_i + 1
               ! ice bottom term
               ztrid   (1, jm, ji) =       - zeta_i(nlay_i, ji) *   zkappa_i(nlay_i-1, ji)   
               ztrid   (2, jm, ji) = 1._wp + zeta_i(nlay_i, ji) * ( zkappa_i(nlay_i-1, ji) + zkappa_i(nlay_i, ji) * zg1 )
               ztrid   (3, jm, ji) = 0._wp
               zindterm(jm,ji)   = ztiold(nlay_i, ji) + zeta_i(nlay_i, ji) *  &
                  &              ( zradab_i(nlay_i, ji) + zkappa_i(nlay_i, ji) * zg1 * t_bo_1d(ji) ) 

               !                               !---------------------!
               IF( h_s_1d(ji) > 0._wp ) THEN   !  snow-covered cells !
                  !                            !---------------------!
                  ! snow interior terms (bottom equation has the same form as the others)
                  DO jm = 3, nlay_s + 1
                     jk = jm - 1
                     ztrid   (1, jm, ji) =       - zeta_s(jk,ji) *   zkappa_s(jk-1,ji)
                     ztrid   (2, jm, ji) = 1._wp + zeta_s(jk,ji) * ( zkappa_s(jk-1,ji) + zkappa_s(jk,ji) )
                     ztrid   (3, jm, ji) =       - zeta_s(jk,ji) *                       zkappa_s(jk,ji)
                     zindterm(jm,ji)   = ztsold(jk,ji) + zeta_s(jk,ji) * zradab_s(jk,ji)
                  END DO
                  
                  ! case of only one layer in the ice (ice equation is altered)
                  IF ( nlay_i == 1 ) THEN
                     ztrid   (3, nlay_s+2, ji) = 0._wp
                     zindterm(nlay_s+2, ji)   = zindterm(nlay_s+2, ji) + zeta_i(1, ji) * zkappa_i(1, ji) * t_bo_1d(ji) 
                  ENDIF
                  
                  jm_min(ji) = 2
                  jm_max(ji) = nlay_i + nlay_s + 1
                  
                  ! first layer of snow equation
                  ztrid   (1, 2, ji) = 0._wp
                  ztrid   (2, 2, ji) = 1._wp + zeta_s(1, ji) * zkappa_s(1, ji)
                  ztrid   (3, 2, ji) =       - zeta_s(1, ji) * zkappa_s(1, ji) 
                  zindterm(2, ji)   = ztsold(1, ji) + zeta_s(1, ji) * ( zradab_s(1, ji) + qcn_ice_1d(ji) ) 
                  
                  !                            !---------------------!
               ELSE                            ! cells without snow  !
                  !                            !---------------------!
                  jm_min(ji) = nlay_s + 2
                  jm_max(ji) = nlay_i + nlay_s + 1
                  
                  ! first layer of ice equation
                  ztrid   (1, jm_min(ji), ji) = 0._wp
                  ztrid   (2, jm_min(ji), ji) = 1._wp + zeta_i(1, ji) * zkappa_i(1, ji)  
                  ztrid   (3, jm_min(ji), ji) =       - zeta_i(1, ji) * zkappa_i(1, ji)
                  zindterm(jm_min(ji), ji)   = ztiold(1, ji) + zeta_i(1, ji) * ( zradab_i(1, ji) + qcn_ice_1d(ji) )
                  
                  ! case of only one layer in the ice (surface & ice equations are altered)
                  IF( nlay_i == 1 ) THEN
                     ztrid   (1, jm_min(ji), ji) = 0._wp
                     ztrid   (2, jm_min(ji), ji) = 1._wp + zeta_i(1, ji) * zkappa_i(1, ji)
                     ztrid   (3, jm_min(ji), ji) = 0._wp
                     zindterm(jm_min(ji), ji)   = ztiold(1, ji) + zeta_i(1, ji) *  &
                        &                                     ( zradab_i(1, ji) + zkappa_i(1, ji) * t_bo_1d(ji) + qcn_ice_1d(ji) )
                  ENDIF
                  
               ENDIF
               !
               zindtbis(jm_min(ji), ji) = zindterm(jm_min(ji), ji)
               zdiagbis(jm_min(ji), ji) = ztrid   (2, jm_min(ji), ji)
               !
            !
            !------------------------------
            ! 8) tridiagonal system solving
            !------------------------------
            ! Solve the tridiagonal system with Gauss elimination method.
            ! Thomas algorithm, from Computational fluid Dynamics, J.D. ANDERSON, McGraw-Hill 1984





            DO jm = jm_min(ji)+1, jm_max(ji)
                  zdiagbis(jm,ji) = ztrid   (2, jm, ji) - ztrid(1, jm, ji) * ztrid   (3, jm-1, ji) / zdiagbis(jm-1,ji)
                  zindtbis(jm,ji) = zindterm(jm,ji)   - ztrid(1, jm,ji) * zindtbis(jm-1,ji)   / zdiagbis(jm-1,ji)
            END DO
            
            ! ice temperatures
               tt_i_1d(nlay_i, ji) = zindtbis(jm_max(ji), ji) / zdiagbis(jm_max(ji), ji)

               DO jm = nlay_i + nlay_s, nlay_s + 2, -1
                  jk = jm - nlay_s - 1
                  tt_i_1d(jk, ji) = ( zindtbis(jm,ji) - ztrid(3, jm, ji) * tt_i_1d(jk+1, ji) ) / zdiagbis(jm,ji)
               END DO
            
            ! snow temperatures      
               IF( h_s_1d(ji) > 0._wp ) THEN
                  tt_s_1d(nlay_s, ji) = ( zindtbis(nlay_s+1, ji) - ztrid(3, nlay_s+1, ji) * tt_i_1d(1, ji) ) / zdiagbis(nlay_s+1, ji)
               ENDIF
            !
            !--------------------------------------------------------------
            ! 9) Has the scheme converged?, end of the iterative procedure
            !--------------------------------------------------------------
            ! check that nowhere it has started to melt
            ! zdti_max is a measure of error, it has to be under zdti_bnd


               DO jk = 1, nlay_s
                     tt_s_1d(jk,ji) = MAX( MIN( tt_s_1d(jk,ji), rt0 ), rt0 - 100._wp )
                     zdti_max = MAX( zdti_max, ABS( tt_s_1d(jk,ji) - ztsb(jk,ji) ) )
               END DO
            
               DO jk = 1, nlay_i
                     ztmelts       = -rTmlt * tsz_i_1d(jk,ji) + rt0 
                     tt_i_1d(jk,ji) =  MAX( MIN( tt_i_1d(jk,ji), ztmelts ), rt0 - 100._wp )
                     zdti_max      =  MAX( zdti_max, ABS( tt_i_1d(jk,ji) - ztib(jk,ji) ) )
               END DO

        ENDIF ! k_jules

        IF(zdti_max.le.zdti_bnd) THEN
          liter(ji) = .FALSE.
          itot = itot - 1
        ENDIF

        ENDIF ! liter
        END DO ! ji
      END DO  ! End of the do while iterative procedure
!need to move this part where all processors are available      
!     IF( ln_icectl .AND. lwp ) THEN
!        WRITE(numout,*) ' zdti_max : ', zdti_max
!        WRITE(numout,*) ' iconv    : ', iconv
!     ENDIF
      
      !
      !-----------------------------
      ! 10) Fluxes at the interfaces
      !-----------------------------
      !
      ! --- calculate conduction fluxes (positive downward)
      DO jk = 1, nlay_s
         t_s_1d(1:npti, jk) = tt_s_1d(jk, 1:npti)
      ENDDO
      DO jk = 1, nlay_i
         t_i_1d(1:npti, jk) = tt_i_1d(jk, 1:npti)
      ENDDO

      DO ji = 1, npti
         !                                ! surface ice conduction flux
         qcn_ice_top_1d(ji) =  -           isnow(ji)   * zkappa_s(0, ji)      * zg1s * ( t_s_1d(ji,1) - t_su_1d(ji) )  &
            &                  - ( 1._wp - isnow(ji) ) * zkappa_i(0, ji)      * zg1  * ( t_i_1d(ji,1) - t_su_1d(ji) )
         !                                ! bottom ice conduction flux
         qcn_ice_bot_1d(ji) =                          - zkappa_i(nlay_i, ji) * zg1  * ( t_bo_1d(ji ) - t_i_1d (ji,nlay_i) )
      END DO
      
      !
      ! --- Diagnose the heat loss due to changing non-solar / conduction flux --- !
      !
      IF( k_jules == np_jules_OFF .OR. k_jules == np_jules_EMULE ) THEN
         !
         DO ji = 1, npti
            hfx_err_dif_1d(ji) = hfx_err_dif_1d(ji) - ( qns_ice_1d(ji) - zqns_ice_b(ji) ) * a_i_1d(ji) 
         END DO
         !
      ELSEIF( k_jules == np_jules_ACTIVE ) THEN
         !
         DO ji = 1, npti
            hfx_err_dif_1d(ji) = hfx_err_dif_1d(ji) - ( qcn_ice_top_1d(ji) - qcn_ice_1d(ji) ) * a_i_1d(ji) 
         END DO
         !
      ENDIF
      
      !
      ! --- Diagnose the heat loss due to non-fully converged temperature solution (should not be above 10-4 W-m2) --- !
      !
      IF( k_jules == np_jules_OFF .OR. k_jules == np_jules_ACTIVE ) THEN
         
         CALL ice_var_enthalpy       
         
         ! zhfx_err = correction on the diagnosed heat flux due to non-convergence of the algorithm used to solve heat equation
         DO ji = 1, npti
            zdq = - zq_ini(ji) + ( SUM( e_i_1d(ji,1:nlay_i) ) * h_i_1d(ji) * r1_nlay_i +  &
               &                   SUM( e_s_1d(ji,1:nlay_s) ) * h_s_1d(ji) * r1_nlay_s )
            
            IF( k_jules == np_jules_OFF ) THEN
               
               IF( t_su_1d(ji) < rt0 ) THEN  ! case T_su < 0degC
                  zhfx_err = ( qns_ice_1d(ji)     + qsr_ice_1d(ji)     - zradtr_i(ji, nlay_i) - qcn_ice_bot_1d(ji)  &
                     &       + zdq * r1_rdtice ) * a_i_1d(ji)
               ELSE                          ! case T_su = 0degC
                  zhfx_err = ( qcn_ice_top_1d(ji) + qtr_ice_top_1d(ji) - zradtr_i(ji, nlay_i) - qcn_ice_bot_1d(ji)  &
                     &       + zdq * r1_rdtice ) * a_i_1d(ji)
               ENDIF
               
            ELSEIF( k_jules == np_jules_ACTIVE ) THEN
            
               zhfx_err    = ( qcn_ice_top_1d(ji) + qtr_ice_top_1d(ji) - zradtr_i(ji, nlay_i) - qcn_ice_bot_1d(ji)  &
                  &          + zdq * r1_rdtice ) * a_i_1d(ji)
            
            ENDIF
            !
            ! total heat sink to be sent to the ocean
            hfx_err_dif_1d(ji) = hfx_err_dif_1d(ji) + zhfx_err
            !
            ! hfx_dif = Heat flux diagnostic of sensible heat used to warm/cool ice in W.m-2   
            hfx_dif_1d(ji) = hfx_dif_1d(ji) - zdq * r1_rdtice * a_i_1d(ji)
            !
         END DO
         !
      ENDIF
      !
      !---------------------------------------------------------------------------------------
      ! 11) Jules coupling: reset inner snow and ice temperatures, update conduction fluxes
      !---------------------------------------------------------------------------------------
      ! effective conductivity and 1st layer temperature (needed by Met Office)
      DO ji = 1, npti
         IF( h_s_1d(ji) > 0.1_wp ) THEN 
            cnd_ice_1d(ji) = 2._wp * zkappa_s(0, ji)
         ELSE
            IF( h_i_1d(ji) > 0.1_wp ) THEN
               cnd_ice_1d(ji) = 2._wp * zkappa_i(0, ji)
            ELSE
               cnd_ice_1d(ji) = 2._wp * ztcond_i(0, ji) * 10._wp
            ENDIF
         ENDIF
         t1_ice_1d(ji) = isnow(ji) * t_s_1d(ji,1) + ( 1._wp - isnow(ji) ) * t_i_1d(ji,1)
      END DO
      !
      IF( k_jules == np_jules_EMULE ) THEN
         ! Restore temperatures to their initial values
         DO jk = 1, nlay_s
            t_s_1d    (1:npti,jk) = ztsold        (jk, 1:npti)
         ENDDO
         DO jk = 1, nlay_i
            t_i_1d    (1:npti,jk) = ztiold        (jk, 1:npti)
         ENDDO
         qcn_ice_1d(1:npti)   = qcn_ice_top_1d(1:npti)
      ENDIF
      !
      ! --- SIMIP diagnostics
      !
      DO ji = 1, npti         
         !--- Snow-ice interfacial temperature (diagnostic SIMIP)
         zfac = rn_cnd_s * zh_i(ji) + ztcond_i(1, ji) * zh_s(ji)
         IF( h_s_1d(ji) >= zhs_min ) THEN
            t_si_1d(ji) = ( rn_cnd_s       * zh_i(ji) * t_s_1d(ji,1) +   &
               &            ztcond_i(1, ji) * zh_s(ji) * t_i_1d(ji,1) ) / MAX( epsi10, zfac )
         ELSE
            t_si_1d(ji) = t_su_1d(ji)
         ENDIF
      END DO
      !
   END SUBROUTINE ice_thd_zdf_BL99

#else
   !!----------------------------------------------------------------------
   !!   Default option       Dummy Module             No SI3 sea-ice model
   !!----------------------------------------------------------------------
#endif

   !!======================================================================
END MODULE icethd_zdf_BL99
